///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoRecoveryManager.cc
// ----------------------
// Cego recovery manager class implementation
//                                                         
// Design and Implementation by Bjoern Lemke               
//
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoRecoveryManager
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

// base includes
#include <lfc/SetT.h>
#include <lfc/Datetime.h>
#include <lfc/Sleeper.h>
#include <lfc/Host.h>
#include <lfc/CommandExecuter.h>

// cego includes
#include "CegoXMLdef.h"
#include "CegoRecoveryManager.h"

#include <string.h>
#include <stdlib.h>

CegoRecoveryManager::CegoRecoveryManager(CegoDistManager *pGTM)
{
    _pGTM = pGTM;
    _pLogger = _pGTM->getDBMng();
    _pDBMng = _pGTM->getDBMng();
    _modId = _pDBMng->getModId("CegoRecoveryManager"); 	    
}

CegoRecoveryManager::~CegoRecoveryManager()
{
}

long CegoRecoveryManager::recoverTableSet(const Chain& tableSet, int pit, bool waitForSync, const Chain& logMngProg)
{   

    _pLogger->log(_modId, Logger::DEBUG, Chain("Recovering tableset ") + tableSet + Chain(" ..."));

    int tabSetId = _pDBMng->getTabSetId(tableSet);

    _pDBMng->setRecoveryMode(tabSetId, CegoDatabaseManager::ON ); 

    if ( waitForSync )
    {
	while ( _pDBMng->getTableSetSyncState(tableSet) != Chain(XML_SYNCHED_VALUE) &&  _pDBMng->getRecoveryMode(tabSetId) == CegoDatabaseManager::ON)
	{
	    _pDBMng->log(_modId, Logger::DEBUG, Chain("Waiting for tableset sync ... "));
	    Sleeper s;
	    s.secSleep(LOGMNG_RECOVERY_DELAY);
	}
    }
    
    if ( _pDBMng->getTableSetSyncState(tableSet) != Chain(XML_SYNCHED_VALUE) )
    {
	Chain msg = Chain("No sync on tableset ") + tableSet + Chain(", recovery failed");
	_pDBMng->setRecoveryMode(tabSetId, CegoDatabaseManager::OFF );
	_pDBMng->log(_modId, Logger::LOGERR, msg);
	throw Exception(EXLOC, msg); 
    }
    
    _pGTM->initLogFiles(tableSet, true);
    _pGTM->initLock(tabSetId);
    _pGTM->regDataFiles(tableSet);
    
    long cplsn;
    
    Chain tsTicketName = _pDBMng->getTSTicket(tableSet);
    
    File tsTicket(tsTicketName);
    if ( tsTicket.exists() ) 
    {
	_pDBMng->log(_modId, Logger::DEBUG, Chain("Backup tableset ticket detected, datafile file recovery required ..."));
	tsTicket.open(File::READ);
	
	XMLSuite xml;
	
	Document *pDoc = new Document;
	pDoc->setAttribute(XML_VERSION_ATTR, XML_VERSION_VALUE);
	
	xml.setDocument(pDoc);    
	xml.setFile(&tsTicket);
	
	xml.parse();
	
	Element *pRoot = pDoc->getRootElement();
	pDoc->setRootElement(0);
	delete pDoc;
	
	_pDBMng->setTableSetInfo(tableSet, pRoot);
	
	cplsn = _pDBMng->getCommittedLSN(tableSet);
	
	_pDBMng->log(_modId, Logger::NOTICE, Chain("Starting datafile recovery for tableset ") + tableSet + Chain(" ..."));
	dataFileRecovery(tableSet, tabSetId, cplsn, logMngProg);
	_pDBMng->log(_modId, Logger::NOTICE, Chain("Datafile recovery for tableset ") + tableSet + Chain(" finished"));
	
	tsTicket.close();

	_pDBMng->log(_modId, Logger::DEBUG, Chain("Removing backup ticket ") + tsTicket.getFileName() + Chain(" ..."));
	
	tsTicket.remove();	
	
    }
    else
    {	
	cplsn = _pDBMng->getCommittedLSN(tableSet);
    }
    
    if ( pit == 0 )
	_pDBMng->log(_modId, Logger::NOTICE, Chain("Starting up-to-crash transaction recovery for tableset ") + tableSet + Chain(" ..."));
    else
    {
	Datetime d(pit);
	_pDBMng->log(_modId, Logger::NOTICE, 
		     Chain("Starting point-in-time transaction recovery to ") + d.asChain() + Chain(" for tableset ") + tableSet + Chain(" ..."));

    }
    
    _pDBMng->setTableSetRunState(tableSet, XML_ONLINE_VALUE);
    
    try
    {
	cplsn = transactionRecovery(tableSet, tabSetId, cplsn, pit, logMngProg);
    }
    catch ( Exception e )
    {
	// Chain msg = e.getBaseMsg();

	_pDBMng->setRecoveryMode(tabSetId, CegoDatabaseManager::OFF );	
	throw Exception(EXLOC, Chain("Recovery failed"), e);
    }
    
    _pDBMng->log(_modId, Logger::DEBUG, Chain("Transaction recovery finished with cplsn = ") + Chain(cplsn));
    
    _pDBMng->setActiveLogFile(tableSet);
    
    _pDBMng->setLSN(tabSetId, cplsn + 1);
    _pDBMng->startLog(tabSetId);
        
    CegoLogRecord logRec;    
    CegoLogRecord lr;
    lr.setAction(CegoLogRecord::LOGREC_SYNC);    
    if ( _pDBMng->logAction(tabSetId, lr) == CegoLogManager::LOG_ERROR )
    {
	throw Exception(EXLOC, "Cannot write sync point after tableset recovery");
    }
        
    _pDBMng->setRecoveryMode(tabSetId, CegoDatabaseManager::OFF );
    
    long lsn = _pDBMng->getLSN(tabSetId);
    _pDBMng->setCommittedLSN(tabSetId, lsn);
    
    _pDBMng->log(_modId, Logger::DEBUG, Chain("Writing sync point ..."));

    _pGTM->writeCheckPoint(tableSet, true, false);
    
    _pDBMng->log(_modId, Logger::NOTICE, Chain("Recovery for tableset ") + tableSet + Chain(" finished"));
    
    return cplsn;
}

void CegoRecoveryManager::dataFileRecovery(const Chain& tableSet, int tabSetId, long cplsn, const Chain& logMngProg)
{

    bool endOfBackup = false;       
    
    while ( endOfBackup == false &&  _pDBMng->getRecoveryMode(tabSetId) == CegoDatabaseManager::ON )
    {
	
	Chain archLogFileName = _pDBMng->getArchiveLogName(tableSet, cplsn + 1);
	
	ListT<Chain> archIdList;
	ListT<Chain> archPathList;
	
	_pDBMng->getArchLogInfo(tabSetId, archIdList, archPathList);
	
	Chain *pArchLogPath = archPathList.First();

	Chain sourceFileName;
	
	bool fileFound = false;

	while ( pArchLogPath  && fileFound == false )
	{
			
	    // check if exists
	    
	    sourceFileName = *pArchLogPath + Chain("/") + archLogFileName;
	    
	    File checkFile(sourceFileName);

	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Checking file ") + sourceFileName);
	    
	    if ( checkFile.exists() )
	    {		    
		fileFound = true;		
	    }
	    else
	    {
		pArchLogPath = archPathList.Next();
	    }
	}

	if ( fileFound )
	{

	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Recovering offline logfile ") + sourceFileName + Chain(" ..."));

	    _pDBMng->setLogFile(tabSetId, sourceFileName, true);
	    _pDBMng->setLSN(tabSetId, cplsn + 1);
	    
	    // recover offline logs
	    cplsn = recoverCurrentDataFileLog(tabSetId, endOfBackup);
	    
	}
	else
	{
	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Waiting for logfile ") + sourceFileName + Chain(" ..."));
	    
	    if ( logMngProg != Chain("") )
	    {
		
		Chain *pArchLogPath = archPathList.First();
		Chain pathString;
		while ( pArchLogPath )
		{
		    pathString += *pArchLogPath;
		    pArchLogPath = archPathList.Next();
		    if ( pArchLogPath )
			pathString += Chain(":");
		}
		
		
		CommandExecuter cmdExe;
		Chain archRestoreProg = _pDBMng->getArchRestoreProg();
		
		Chain restoreCmd = archRestoreProg + Chain(" ") + archLogFileName + Chain(" ") + pathString; 
		
		_pDBMng->log(_modId, Logger::NOTICE, Chain("Triggering external log manager with <") + Chain(restoreCmd) + Chain(">"));
		int retCode = cmdExe.execute(restoreCmd, LOGMNG_RECOVERY_DELAY);
		_pDBMng->log(_modId, Logger::NOTICE, Chain("External log manager returned : <") + Chain(retCode) + Chain(">"));
		if ( retCode == 0 )
		{
		    // file could be provided
		}
		else if ( retCode == 1 )
		{
		    endOfBackup = true;
		}
		else // ( retCode > 1 )
		{
		    Chain msg("External log maager failed");
		    throw Exception(EXLOC, msg);			    
		}
	    }
	    else
	    {
		Sleeper s;
		s.secSleep(LOGMNG_RECOVERY_DELAY);
	    }
	}
    }
}

long CegoRecoveryManager::transactionRecovery(const Chain& tableSet, int tabSetId, long cplsn, int pit, const Chain& logMngProg)
{
    bool endOfRecovery = false;    
    Chain sourceFileName;
    bool pitReached = false;
    int ts = 0;
	    
    while ( _pDBMng->getRecoveryMode(tabSetId) == CegoDatabaseManager::ON && pitReached == false && endOfRecovery == false)
    {

	bool fileFound = false;   

	while ( fileFound == false && endOfRecovery == false &&  _pDBMng->getRecoveryMode(tabSetId) == CegoDatabaseManager::ON )
	{
	    
	    Chain archLogFileName = _pDBMng->getArchiveLogName(tableSet, cplsn + 1);
	    
	    ListT<Chain> archIdList;
	    ListT<Chain> archPathList;
	    
	    _pDBMng->getArchLogInfo(tabSetId, archIdList, archPathList);
	    
	    Chain *pArchLogPath = archPathList.First();
	    
	    while ( pArchLogPath  && fileFound == false)
	    {
		
		// check if exists
		
		sourceFileName = *pArchLogPath + Chain("/") + archLogFileName;
		
		File checkFile(sourceFileName);
		
		_pDBMng->log(_modId, Logger::NOTICE, Chain("Checking logfile ") + sourceFileName + Chain(" ..."));
		
		if ( checkFile.exists() )
		{		    
		    fileFound = true;
		    
		}
		else
		{
		    pArchLogPath = archPathList.Next();
		}
	    }
	    if ( fileFound == false )
	    {
		if ( logMngProg != Chain("") )
		{	
		    // if no external manager is defined, we finish recovery here
		    if ( logMngProg == Chain(LOGMNG_NONEPROG) )
		    {
			endOfRecovery = true;
		    }
		    else
		    {
			_pDBMng->log(_modId, Logger::NOTICE, Chain("Waiting for logfile ") + archLogFileName + Chain(" ..."));
			
			Chain *pArchLogPath = archPathList.First();
			Chain pathString;
			while ( pArchLogPath )
			{
			    pathString += *pArchLogPath;
			    pArchLogPath = archPathList.Next();
			    if ( pArchLogPath )
				pathString += Chain(":");
			}
			
			CommandExecuter cmdExe;
			Chain archRestoreProg = _pDBMng->getArchRestoreProg();


			Chain restoreCmd = archRestoreProg + Chain(" ") + archLogFileName + Chain(" ") + pathString; 

			
			_pDBMng->log(_modId, Logger::NOTICE, Chain("Triggering external log manager with <") + Chain(restoreCmd) + Chain(">"));
			int retCode = cmdExe.execute(restoreCmd, LOGMNG_RECOVERY_DELAY);
			_pDBMng->log(_modId, Logger::NOTICE, Chain("External log manager returned : <") + Chain(retCode) + Chain(">"));
			if ( retCode == 0 )
			{
			    // file could be provided
			}
			else if ( retCode == 1 )
			{
			    endOfRecovery = true;
			}
			else // ( retCode > 1 )
			{
			    Chain msg("External log maager failed");
			    throw Exception(EXLOC, msg);			    
			}		     
		    }
		}
		else
		{
		    Sleeper s;
		    s.secSleep(LOGMNG_RECOVERY_DELAY);
		}
	    }
	}
	bool isRecovered = false;
	while ( fileFound && ! isRecovered  && _pDBMng->getRecoveryMode(tabSetId) == CegoDatabaseManager::ON )
	{
	
	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Recovering offline logfile ") + sourceFileName + Chain(" ..."));
	    _pDBMng->setLogFile(tabSetId, sourceFileName, true);
	    
	    _pDBMng->setLSN(tabSetId, cplsn + 1);

	    long lsn;
	    	    
	    pitReached = recoverCurrentTransactionLog(tabSetId, pit, lsn, ts);

	    if ( lsn > 0 )
	    {
		cplsn = lsn;
		_pGTM->writeCheckPoint(tableSet, false, false);
		isRecovered=true;
	    }
	    else if ( pitReached == false ) 
	    {
		_pDBMng->log(_modId, Logger::NOTICE, Chain("Incomplete offline logfile ") + sourceFileName + Chain(" detected, waiting ..."));
		Sleeper s;
		s.secSleep(LOGMNG_RECOVERY_DELAY);
	    }
	}
    }


    if ( _pDBMng->getRecoveryMode(tabSetId) != CegoDatabaseManager::ON )
    {
	Chain msg("Transaction recovery aborted");
	throw Exception(EXLOC, msg);
    }


    return cplsn;
}

int CegoRecoveryManager::recoverCurrentDataFileLog(int tabSetId, bool& endOfBackup)
{
    
    long lsn=0;

    _pDBMng->seekToStart(tabSetId);

    _pDBMng->log(_modId, Logger::NOTICE, Chain("Recovering logfile ... "));

    CegoLockHandler* pLockHandle = _pGTM->getLockHandle();

    // now recover to last lsn in log
    
    endOfBackup = false;
    CegoLogRecord logRec;
    while ( _pDBMng->logRead(tabSetId, logRec) && ! endOfBackup )
    {
	
	if ( logRec.getLSN() == _pDBMng->getLSN(tabSetId)  )
	{
	    
	    lsn=logRec.getLSN();
	    	    
	    switch ( logRec.getAction() ) 
	    {
		
	    case CegoLogRecord::LOGREC_BUPAGE:
		
		_pDBMng->log(_modId, Logger::DEBUG, Chain("Writing page (") + Chain(logRec.getFileId()) + Chain(",") + Chain(logRec.getPageId()) + Chain(")"));
		_pDBMng->writePage(logRec.getFileId(),logRec.getPageId(), 0, logRec.getData(), pLockHandle ); 
		break;
		
	    case CegoLogRecord::LOGREC_BUFBM:
	    {
		_pDBMng->log(_modId, Logger::DEBUG, Chain("Writing fbm for fileId ") + Chain(logRec.getFileId()));
		_pDBMng->writeFBM(logRec.getFileId(), (unsigned*)logRec.getData(), pLockHandle ); 
		break;
	    }
	    case CegoLogRecord::LOGREC_BUFIN:
	    {
		_pDBMng->log(_modId, Logger::NOTICE, Chain("Detected end of backup"));
		endOfBackup=true;
		break;
	    }
	    default:
		break;
	    }

	    _pDBMng->setLSN(tabSetId, lsn + 1);
	}
	else if ( logRec.getLSN() < _pDBMng->getLSN(tabSetId)  )
	{
	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Ignoring lsn ") + Chain(logRec.getLSN()) + Chain(" ( expected ") + Chain(_pDBMng->getLSN(tabSetId)) + Chain(")"));	    
	}
	else
	{
	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Skipping logfile lsn ") + Chain(logRec.getLSN()) + Chain(" ( expected ") + Chain(_pDBMng->getLSN(tabSetId)) + Chain(")"));	    
	    endOfBackup=true;

	}

	if ( logRec.getData() )
	    free ( logRec.getData() ) ;
	
    }
    
    _pDBMng->log(_modId, Logger::NOTICE, Chain("Datafile recovery finished with lsn ") + Chain(lsn));
   
    return lsn;
    
}

bool CegoRecoveryManager::recoverCurrentTransactionLog(int tabSetId, int pit, long& lsn, int& ts)
{
    
    lsn=0;
    bool pitReached = false;

    _pDBMng->seekToStart(tabSetId);

    _pDBMng->log(_modId, Logger::NOTICE, Chain("Recovering logfile ..."));

    // now recover to last lsn in log

    SetT<Chain> invalidateSet;

    CegoLogRecord logRec;
    while ( _pDBMng->logRead(tabSetId, logRec) && pitReached == false )
    {
	if ( pit > 0 && logRec.getTS() > pit )
	{
	    Datetime d(logRec.getTS());
	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Required point in time reached, Ignoring logentry  ") + Chain(logRec.getLSN()) 
			 + Chain(" ( ") + d.asChain() + Chain(" ) and higher"));
	    
	    pitReached = true;
	}
	else if ( logRec.getTS() <  ts )
	{
	    _pDBMng->log(_modId, Logger::NOTICE, Chain("Old log detected, ignoring logentry ") + Chain(logRec.getLSN()) + Chain(" and higher"));
	    pitReached = true;
	}
	else
	{
	    if ( logRec.getLSN() == _pDBMng->getLSN(tabSetId)  )
	    {

		lsn=logRec.getLSN();
		ts=logRec.getTS();
		int tid=logRec.getTID();

		_pDBMng->log(_modId, Logger::DEBUG, Chain("Recovering lsn ") + Chain(logRec.getLSN()) + Chain(" ..."));
		
		// here comes the recovery
		
		switch ( logRec.getAction() ) 
		{
		    
		case CegoLogRecord::LOGREC_CREATE:
		    
		    switch ( logRec.getObjType() )
		    {
			
		    case CegoObject::TABLE:
		    case CegoObject::SYSTEM:
		    {
			CegoTableObject to;
			to.decode(logRec.getData());
			_pGTM->createDataTable(tabSetId, logRec.getObjName(), logRec.getObjType(), to.getSchema());
			break;
		    }
		    case CegoObject::INDEX:
		    case CegoObject::PINDEX:
		    case CegoObject::UINDEX:
		    {
			CegoTableObject to;
			to.decode(logRec.getData());
			_pGTM->createIndexTable(tabSetId, logRec.getObjName(), to.getTabName(), to.getSchema(), logRec.getObjType());
			break;
		    }
		    case CegoObject::VIEW:
		    {
			CegoViewObject vo;
			vo.decode(logRec.getData());
			_pGTM->createViewObject(vo);
			break;		    
		    }
		    case CegoObject::PROCEDURE:
		    {
			CegoProcObject po;
			po.decode(logRec.getData());
			_pGTM->createProcObject(po);
			break;		    
		    }
		    case CegoObject::FKEY:
		    {
			CegoKeyObject ko;
			ko.decode(logRec.getData());
			_pGTM->createKeyObject(ko);
			break;
		    }
		    default:		    
			break;
		    }
		    break;
		case CegoLogRecord::LOGREC_DROP:
		{
		    _pGTM->removeObject(tabSetId, logRec.getObjName(), logRec.getObjType());
		    break;
		}
		case CegoLogRecord::LOGREC_ALTER:
		{
		    CegoTableObject aoe;
		    aoe.decode(logRec.getData());
		    _pGTM->alterTableObject(tabSetId, logRec.getObjName(), logRec.getObjType(), aoe);
		    break;
		}
		case CegoLogRecord::LOGREC_RENAME:
		{
		    Chain newObjName(logRec.getData(), logRec.getDataLen());
		    _pGTM->renameObject(tabSetId, logRec.getObjName(), logRec.getObjType(), newObjName );
		    break;   
		}
		case CegoLogRecord::LOGREC_BEGIN:
		{
		    // _pGTM->beginTransaction(tabSetId);
		    break;
		}
		case CegoLogRecord::LOGREC_COMMIT:
		{
		    _pGTM->setTID(tabSetId, tid);		    
		    _pGTM->commitTransaction(tabSetId);
		    break;
		}
		case CegoLogRecord::LOGREC_ABORT:
		{
		    _pGTM->setTID(tabSetId, tid);
		    _pGTM->rollbackTransaction(tabSetId);
		    break;
		}
		case CegoLogRecord::LOGREC_INSERT:
		{
		    
		    CegoTableObject oe;
		    _pGTM->getObject(tabSetId, logRec.getObjName(), CegoObject::TABLE, oe);

		    if ( invalidateSet.Find(logRec.getObjName()) == 0 )
		    {
			_pDBMng->log(_modId, Logger::NOTICE, Chain("Invalidating index for table ") + logRec.getObjName());		    
			_pGTM->invalidateIndexForTable(tabSetId, logRec.getObjName());
			invalidateSet.Insert(logRec.getObjName());
		    }
		    ListT<CegoField> fvl = oe.getSchema();
		    
		    int rectid = 0;
		    int tastep = 0;
		    CegoTupleState ts;
		    ListT<CegoBlob> blobList;
		    CegoQueryHelper qh;
		    qh.decodeFVL(fvl, blobList, logRec.getData(), logRec.getDataLen(), rectid, tastep, ts);
		    
		    CegoBlob *pBlob = blobList.First();
		    CegoField *pF = fvl.First();
		    while ( pBlob && pF )
		    {
	    	    
			int fileId;
			int pageId;
			
			_pGTM->putBlobData(tabSetId, pBlob->getBufPtr(), pBlob->getSize(), fileId, pageId);
			
			while ( pF->getValue().getType() != BLOB_TYPE && pF)
			    pF = fvl.Next();
			
			if ( pF )
			{
			    Chain blobRef = Chain("[") + Chain(fileId) + Chain(",") + Chain(pageId) + Chain("]");
			    CegoFieldValue fv(BLOB_TYPE, blobRef);
			    pF->setValue(fv);
			}
			else
			{
			    throw Exception(EXLOC, "Cannot get blob reference");
			}
			
			free ( pBlob->getBufPtr() );
			pBlob = blobList.Next();

		    }
		    
		    bool lastPage = true;
		    CegoDataPointer dp;

		    _pGTM->setTID(tabSetId, tid);
		    _pGTM->insertDataTable(oe, fvl, dp, lastPage);
		    
		    break;
		}
		case CegoLogRecord::LOGREC_DELETE:
		{
		  
		    CegoPredDesc *pPred = 0;	    
		    CegoQueryHelper qh;
		    Chain tableAlias;

		    qh.decodeDelRec(tableAlias, pPred, logRec.getData(), logRec.getDataLen(), _pGTM);

		    _delRecList.Insert(DeleteRecord(tid, tableAlias, pPred));
		    
		    break;
		}
		case CegoLogRecord::LOGREC_DELETE_COMMIT:
		{
 
		    DeleteRecord* delRecord = _delRecList.Find(DeleteRecord(tid));
		    
		    if ( delRecord )
		    {

			CegoTableObject oe;
			_pGTM->getObject(tabSetId, logRec.getObjName(), CegoObject::TABLE, oe);
			
			if ( invalidateSet.Find(logRec.getObjName()) == 0 )
			{
			    _pDBMng->log(_modId, Logger::NOTICE, Chain("Invalidating index for table ") + logRec.getObjName());		    
			    _pGTM->invalidateIndexForTable(tabSetId, logRec.getObjName());
			    invalidateSet.Insert(logRec.getObjName());
			}
			
			_pGTM->setTID(tabSetId, tid);
			_pGTM->deleteDataTable(oe, delRecord->getPred(), 0);		    

			delRecord->clean();
			_delRecList.Remove(DeleteRecord(tid));

			
		    }
		    else
		    {
			_pDBMng->log(_modId, Logger::NOTICE, Chain("Cannot find delete record for transaction ") + Chain(tid));
			
		    }
		    break;
		}
		case CegoLogRecord::LOGREC_UPDATE:
		{
		    
		    ListT<CegoField> updList;
		    ListT<CegoExpr*> exprList;

		    CegoPredDesc *pPred = 0;		    
		    CegoQueryHelper qh;
		    Chain tableAlias;
		    qh.decodeUpdRec(tableAlias, pPred, updList, exprList, logRec.getData(), logRec.getDataLen(), _pGTM);

		    _updRecList.Insert(UpdateRecord(tid, tableAlias, pPred, updList, exprList));

		    break;
		}
		case CegoLogRecord::LOGREC_UPDATE_COMMIT:
		{

		    UpdateRecord* updRecord = _updRecList.Find(UpdateRecord(tid));
		    
		    if ( updRecord )
		    {
		    
			if ( invalidateSet.Find(logRec.getObjName()) == 0 )
			{
			    _pDBMng->log(_modId, Logger::NOTICE, Chain("Invalidating index for table ") + logRec.getObjName());		    
			    _pGTM->invalidateIndexForTable(tabSetId, logRec.getObjName());
			    invalidateSet.Insert(logRec.getObjName());
			}

			_pGTM->setTID(tabSetId,tid); 
			_pGTM->updateDataTable(tabSetId, logRec.getObjName(),
					       updRecord->getTable(), 
					       updRecord->getPred(), 
					       updRecord->getUpdateList(), 
					       updRecord->getExprList(), 0);
			
			updRecord->clean();
			_updRecList.Remove(UpdateRecord(tid));
		    }
		    else
		    {
			_pDBMng->log(_modId, Logger::NOTICE, Chain("Cannot find update record for transaction ") + Chain(tid));			
		    }
		    break;
		}
		case CegoLogRecord::LOGREC_ADDCOUNTER:
		{
		    Chain counterName ( logRec.getData() );
		    long initValue = 0;
		    _pDBMng->addCounter(tabSetId, counterName, initValue);
		    break;
		}
		case CegoLogRecord::LOGREC_DELCOUNTER:
		{
		    Chain counterName ( logRec.getData() );
		    long initValue = 0;
		    _pDBMng->removeCounter(tabSetId, counterName);
		    break;
		}
		default:
		    break;
		}
		_pDBMng->setLSN(tabSetId, lsn + 1);
	    
	    }
	    else if ( logRec.getLSN() < _pDBMng->getLSN(tabSetId)  )
	    {
		_pDBMng->log(_modId, Logger::NOTICE, Chain("Ignoring lsn ") + Chain(logRec.getLSN()) 
			     + Chain(" ( expected ") + Chain(_pDBMng->getLSN(tabSetId)) + Chain(")"));
	    }
	    else
	    {
		_pDBMng->log(_modId, Logger::NOTICE, Chain("Skipping logfile lsn ") + Chain(logRec.getLSN()) 
			     + Chain(" ( expected ") + Chain(_pDBMng->getLSN(tabSetId)) + Chain(")"));
		pitReached = true;
		
	    }
	}
		    
	if ( logRec.getData() )
	    free ( logRec.getData() ) ;

    }
    
    _pDBMng->log(_modId, Logger::NOTICE, Chain("Recovery finished with lsn ") + Chain(lsn));
			 
    return pitReached;
    
}


